require("RYUKU_StarvingZombies_SmellFuncs")

local szVars

--------------------
-- Body Discovery --
--------------------

local deadBodies = {} ---@type zombie.iso.objects.IsoDeadBody[]

---@param isoDeadBody zombie.iso.objects.IsoDeadBody
local function onDeadBodySpawn(isoDeadBody)
    if not isoDeadBody:isSkeleton() and not isoDeadBody:isFakeDead()
    then
        table.insert(deadBodies, isoDeadBody)
    end
end

Events.OnDeadBodySpawn.Add(onDeadBodySpawn)

---@param isoGridSquare zombie.iso.IsoGridSquare
local function loadGridsquare(isoGridSquare)
    local body
    local bodies = isoGridSquare:getDeadBodys()
    for i = 0, bodies:size() - 1 do
        body = bodies:get(i)
        if not body:isSkeleton() and not body:isFakeDead()
        then
            table.insert(deadBodies, body)
        end
    end
end

Events.LoadGridsquare.Add(loadGridsquare)

--------------------
-- isExistInWorld --
--------------------

---@param isoDeadBody zombie.iso.objects.IsoDeadBody
---@return boolean
local function isExistInWorld(isoDeadBody)
    -- existing function doesn't work for dead bodies
    local square = isoDeadBody:getSquare()
    if square ~= nil then
        local bodies = square:getDeadBodys()
        return bodies ~= nil and bodies:size() > 0 and bodies:contains(isoDeadBody)
    end

    return false
end

---------------------
-- Create Skeleton --
---------------------

---@param isoDeadBody zombie.iso.objects.IsoDeadBody
---@param bodyData table
local function createSkeleton(isoDeadBody, bodyData)
    local body
    if isoDeadBody:isAnimal() then
        local animalType = bodyData.AnimalType
        local animalBreed = bodyData.AnimalBreed
        local animalDef = AnimalPartsDefinitions.animals[animalType .. animalBreed]
        if animalDef ~= nil and not animalDef.noSkeleton then
            local animal = IsoAnimal.new(
                getCell(),
                isoDeadBody:getX(),
                isoDeadBody:getY(),
                isoDeadBody:getZ(),
                animalType,
                animalBreed
            )

            animal:setDir(isoDeadBody:getDir())

            body = IsoDeadBody.new(animal, false)
            bodyData.skeleton = "true"
            body:setX(isoDeadBody:getX())
            body:setY(isoDeadBody:getY())
            body:setZ(isoDeadBody:getZ())
            body:setModData(bodyData)
            animal:remove()
        end
    else
        local zombie = addZombiesInOutfit(
            isoDeadBody:getX(), isoDeadBody:getY(), isoDeadBody:getZ(),
            1, -- count
            nil, -- outfit
            0, -- setFemaleEtc
            false, -- crawler
            isoDeadBody:isFallOnFront(),
            false, -- fakeDead
            true, -- knockedDown
            false, -- setInvulnerable
            false, -- setSitAgainstWall
            1 -- setHealth
        )

        -- zombie creation failure check
        -- apparently this can happen when:
        --   zombies are disabled + not in debug
        --   addZombiesInOutfit fails to get the square for some reason
        --   VirtualZombieManager.createRealZombieAlways fails for whatever reason
        -- just delete the body, and move on
        if zombie:size() == 0 then
            bodyData.szSkeleton = true
            isoDeadBody:removeFromWorld()
            isoDeadBody:removeFromSquare()

            for k, v in pairs(deadBodies) do
                if v == isoDeadBody then
                    table.remove(deadBodies, k)
                    return
                end
            end

            return
        end

        zombie = zombie:get(0)
        zombie:setSkeleton(true)
        zombie:setCurrent(isoDeadBody:getSquare())
        zombie:setDir(isoDeadBody:getDir())
        zombie:getInventory():removeAllItems()

        body = IsoDeadBody.new(zombie, false)

        -- stop burnt skeleton appearing
        body:getHumanVisual():setSkinTextureIndex(ZombRand(2) + 1)

        -- add items to inventory
        if szVars.AllowWornAttachedItems then
            body:setWornItems(isoDeadBody:getWornItems())
            body:setAttachedItems(isoDeadBody:getAttachedItems())
        end

        body:setContainer(isoDeadBody:getContainer())

        -- remove items based on chance
        local x = 0
        local container = body:getContainer()
        local allItems = container:getItems()
        while x < allItems:size() do
            if ZombRand(100) < szVars.ItemDestroyChance then
                container:DoRemoveItem(allItems:get(x))
            else
                x = x + 1
            end
        end

        zombie:removeFromWorld()
    end

    if body ~= nil then
        body:invalidateCorpse()
        body:setInvalidateNextRender(true)
    end

    bodyData.szSkeleton = true
    isoDeadBody:removeFromWorld()
    isoDeadBody:removeFromSquare()

    for k, v in pairs(deadBodies) do
        if v == isoDeadBody then
            table.remove(deadBodies, k)
            return
        end
    end
end

-------------
-- Path To --
-------------

---@param isoZombie zombie.characters.IsoZombie
---@param zombieVec SZVector
---@param dir SZVector
---@param dist number
---@param targetZ number
local function doPath(isoZombie, zombieVec, dir, dist, targetZ)
    isoZombie:pathToSound(
        zombieVec.x + dir.x * dist,
        zombieVec.y + dir.y * dist,
        zombieVec.z + dir.z * dist
    )

    if zombieVec.z ~= targetZ then isoZombie:setVariable("bPathfind", true) end
end

--------------------------
--- Update Eat Progress --
--------------------------

---@param isoZombie zombie.characters.IsoZombie
---@param isoDeadBody zombie.iso.objects.IsoDeadBody?
---@return boolean
local function updateEatProgress(isoZombie, isoDeadBody)
    if isoDeadBody == nil then return false end

    local bodyData = isoDeadBody:getModData()
    local time = getTimeInMillis()
    local progress = bodyData.szProgress
    if not progress then
        bodyData.szProgress = { time = time, perc = 0 }
    elseif progress.perc >= 100 then
        if not bodyData.szSkeleton then createSkeleton(isoDeadBody, bodyData) end
        isoZombie:setEatBodyTarget(nil, true)
        isoZombie:getModData().szBody = nil
    else
        local eatTime = szVars.BaseEatTime
        if isoDeadBody:isAnimal() then eatTime = eatTime * szBodyMultiplier[bodyData.AnimalType]
        else eatTime = eatTime * szBodyMultiplier.human end
        eatTime = 100 / eatTime

        local deltaPerc = math.min(1, (time - progress.time) / 1000) * eatTime
        deltaPerc = deltaPerc * getGameTime():getTrueMultiplier()
        progress.perc = progress.perc + deltaPerc
        progress.time = time
    end

    return true
end

----------------
-- Find / Eat --
----------------

local bodyIdx
local bodies

---@param isoZombie zombie.characters.IsoZombie
local function onZombieUpdate(isoZombie)
    -- Bandits compatibility
    if isoZombie:getVariableBoolean("Bandit") or isoZombie:isUseless() then return end

    if not szVars then szVars = SandboxVars.StarvingZombies end
    local modData = isoZombie:getModData()

    -- if zombie is thumping / has a target, clean up
    if isoZombie:getThumpTarget() ~= nil or isoZombie:getTarget() ~= nil then
        modData.szBody = nil
        modData.szBodies = nil
        modData.szPlayer = nil
        isoZombie:setEatBodyTarget(nil, true)
        return
    end

    -- when a zombie is eating, update eaten percentage and spawn skeleton if necessary
    if updateEatProgress(isoZombie, isoZombie:getEatBodyTarget()) then return end

    -- handle target
    if modData.szBody then
        local isoDeadBody = modData.szBody
        if isoDeadBody:isSkeleton() or isoDeadBody:getModData().szSkeleton
            or not isExistInWorld(isoDeadBody)
            or isoDeadBody:getEatingZombies():size() >= 3
            or not szCanSmellBody(isoZombie, isoDeadBody)
        then
            modData.szBody = nil
            isoZombie:setVariable("bPathfind", false)
            isoZombie:setVariable("bMoving", false)
        else
            local zombieVec = SZVector:new(isoZombie)
            local bodyVec = SZVector:new(isoDeadBody)
            local bodyDir = bodyVec:sub(zombieVec)
            local bodyDist = bodyDir:mag()
            if bodyDist > 0.8 then
                -- wait for zombie to be idle
                if isoZombie:getCurrentStateName() == "ZombieIdleState" then
                    -- check for pathing issues
                    local time = getTimeInMillis()
                    local lastPath = modData.szLastPath
                    if lastPath ~= nil and time - lastPath < 250 then
                        modData.szBody = nil
                    else
                        bodyDir:normalise()
                        doPath(isoZombie, zombieVec, bodyDir, bodyDist, bodyVec.z)
                        modData.szLastPath = time
                    end
                end
            else
                isoZombie:setVariable("bPathfind", false)
                isoZombie:setVariable("bMoving", false)
                isoZombie:setEatBodyTarget(isoDeadBody, true)
            end
        end

        return
    elseif modData.szPlayer then
        local isoPlayer = modData.szPlayer
        if not szCanSmellPlayer(isoZombie, isoPlayer) then
            modData.szPlayer = nil
        elseif isoZombie:getCurrentStateName() == "ZombieIdleState" then
            local zombieVec = SZVector:new(isoZombie)
            local playerVec = SZVector:new(isoPlayer)
            local playerDir = playerVec:sub(zombieVec)
            local playerDist = playerDir:mag()
            playerDir:normalise()
            doPath(isoZombie, zombieVec, playerDir, playerDist, playerVec.z)
        end

        return
    end

    -- see if can smell player
    if szVars.CanSmellPlayer and szCanSmellPlayer(isoZombie, getPlayer()) then
        modData.szPlayer = getPlayer()
        return
    end

    -- find bodies to eat
    bodyIdx = modData.szIdx or 0
    bodyIdx = bodyIdx + 1
    if bodyIdx > #deadBodies then
        bodyIdx = 0
        bodies = modData.szBodies
        if bodies ~= nil and #bodies > 0 then
            local idx = ZombRand(#bodies) + 1
            modData.szBody = bodies[idx]
            table.remove(bodies, idx)
            return
        end
    end

    modData.szIdx = bodyIdx
    if #deadBodies == 0 or bodyIdx == 0 then return end
    local isoDeadBody = deadBodies[bodyIdx]

    if isoDeadBody:isSkeleton() or not isExistInWorld(isoDeadBody) then
        table.remove(deadBodies, bodyIdx)
        return
    end

    -- limit to three zombies eating a corpse
    if isoDeadBody:getEatingZombies():size() >= 3 then return end

    -- only select bodies the zombie can smell
    if not szCanSmellBody(isoZombie, isoDeadBody) then return end

    if modData.szBodies == nil then
        modData.szBodies = {} ---@type zombie.iso.objects.IsoDeadBody[]
    end

    table.insert(modData.szBodies, isoDeadBody)
end

Events.OnZombieUpdate.Add(onZombieUpdate)
